/******************************************************************
 * Qiang Zheng,
 * ECS Dept, Syracuse Univ. Nov. 10, 1999
 * for CSE 791 class, fall, 1999, instructor: Steve Chapin
 * O.S.: Linux RedHat 6.0
 * Libraries: Lam MPI library and pnm library for image processing
 ******************************************************************/

#include <stdio.h>
#include <mpi.h>
#include <pnm.h>

/*
 * Data types will be used which are defined in PNM lib:
 * typedef ... xel;              three xelvals
 * typedef ... xelval;           range from 0 to PNM_MAXMAXVAL
 * #define PNM_MAXMAXVAL ...     normally it is 1
 * extern xelval pnm_pbmmaxval;
 *
 * Input and Output functions available in pnm library:
 * Input: xel** pnm_readpnm( FILE* fp, int* colsP, int* rowsP
 *                           , xelval* maxvalP, int* formatP)
 * Output: void pnm_writepnm( FILE* fp, xel** xels, int cols
 *                            , int rows, xelval maxval, int format
 *                            , int forceplain )
 */

#define	WORKTAG	1
#define DIETAG	0

#define UPLIMIT 10
#define LOLIMIT 0

#define MAXLINE 100
/* the max length of a line */

main( int argc, char *argv[] )
{
    int myRank;

    MPI_Init( &argc, &argv ); /* Initialize MPI environment */
    pnm_init( &argc, &argv ); /* routine function in pnm library */

    MPI_Comm_rank( MPI_COMM_WORLD, &myRank ); 

    if( myRank == 0 ){
        master();             /* do the master job */
    } else {
        slave( myRank );              /* do the slave job */
    }

    MPI_Finalize();           /* cleanup MPI */
}

/*
 * The master function:
 * 1) reads image files,
 * 2) partition the image matrix into blocks,
 * 3) assign them to different slave processes,
 * 4) and write the result to image files
 * The schema is map one or several rows into one process
 * so that the boundaries between two adjecent blocks are just
 * two rows, the last row of the previous block and the first
 * row of the next block.
 */
master( )
{
    char *inFileName, *outFileName, line[ MAXLINE ], moreFile; 
    FILE *inFile, *outFile;

    int iterTimes, ntasks, rank;
    int colNum, rowNum, blockSize, imageFormat, i;
    xel **image, *buf, *lastRow, **newBuf;
    /* since images are matrices, they can be regarded as rectangular shape */

    MPI_Status  status;

    /* Read the user input about the filename and the iteration times here */
    /* do {
     printf("Please specify the file name: \n");
     printf("Please specify the iteration times: \n");
     printf("Smoothing input file: %s %d times \n", inFileName, iterTimes)
     printf("Any more files? (y/n) \n");
     moreFile = getChar();
    } while ( moreFile == 'y' || moreFile == 'Y' ) ;
     */

    MPI_Comm_size( MPI_COMM_WORLD, &ntasks );

printf("Master: number of nodes = %d \n", ntasks);

    image = pnm_readpnm( inFile, &colNum, &rowNum, &maxval, &imageFormat);
    buf = image[ 0 ];
    lastRow = image[ rowNum-1 ]; /* offset of the last row */

    rowsPerProc = rowNum/ntasks ;
    blockSize = colNum*rowsPerProc;
    /* size of the block of data each slave process gets */
    lastBlockSize = colNum * ( rowNum-rowsPerProc*(ntasks-2 ) )
    /* size of the last data block the last slave process gets */

    for ( rank=1; rank<ntasks; rank++ ) {

      /* 
       * prototype :
       * MPI_Send( buffer, count, datatype, destination, tag, MPI_COM_WORLD );
       */

      /* To tell the slave how many iteration times are needed */
      MPI_Send( &iterTimes, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
      /* To tell the slave how many columns in each row */
      MPI_Send( &colNum, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );

      /* To tell slaves the size of the blocks will be sent to them */
      if( rank==ntasks-1 ) {
        MPI_Send( lastBlockSize, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
      } else {
        MPI_Send( blockSize, 1, MPI_INT rank, WORKTAG, MPI_COMM_WORLD );
      }

      /*
       * To send the upper boundary to slave before computing 
       * If it is the first slave, copy the first row to its upper boundary
       */
      /* if ( rank == 1 ) { 
       * MPI_Send( buf, colNum*sizeof(xel), MPI_BYTE,
       *   rank, WORKTAG, MPI_COMM_WORLD );
       * } else {
       * MPI_Send( buf-colNum, colNum*sizeof(xel), MPI_BYTE,
       *   rank, WORKTAG, MPI_COMM_WORLD );
       * }
       */

      /*
       * To send the data block to slaves
       * if it is the last slave, the block size may be less than
       * the ragular block size
       */
      if ( rank==ntasks-1 ) {
        MPI_Send( buf, lastBlockSize*sizeof(xel), MPI_BYTE, rank,
          WORKTAG, MPI_COMM_WORLD );
        /*
         * The last row is the lower boundary of the last block
         * here we don't send it 
         */
        /* MPI_Send( lastRow, colNum*sizeof(xel), MPI_BYTE, rank,
			WORKTAG, MPI_COMM_WORLD ); */
      } else {
        MPI_Send( buf, blockSize*sizeof(xel), MPI_BYTE, rank,
          WORKTAG, MPI_COMM_WORLD );
        buf += blockSize;
        /* MPI_Send( buf, colNum*sizeof(xel), MPI_BYTE, rank,
          WORKTAG, MPI_COMM_WORLD ); */
      }

     }

    for( rank=1; rank<ntasks; rank++ ) {
        MPI_Recv( &result, 1, MPI_BYTE, MPI_ANY_SOURCE
            , MPI_ANY_TAG, MPI_COMM_WORLD, &status );
        /* re-organize the data collected from slaves */

        MPI_Send( 0, 0, MPI_INT, rank, DIETAG, MPI_COMM_WORLD );
    }
/* free memory */
}

/*
 * the slave funtion smooth the block it gets from the master
 * and return the result, if necessary, it will exchange the boundrary
 * data with its neighbors
 */ 
slave( int rankNum )
{
    int         blockSize = 0, iterTimes, colNum, rowNum, i, j,
                myRank, ntasks;
    xel         *work, *swapBlock, *temp0, *temp1, *upBound, *lowBound
                *prevRow, *nextRow, *lastRow, *tempP;
    /* pointer `work' points to the block of data to be computed
     * including the two boundaries,
     * swapBlock points to another copy
     * upBound points to the first row of work buffer, e.g. upper boundary
     * lowBound points to the last row of work buffer, e.g. lower boundary
     */
    MPI_Status  status;

    myRank = rankNum;
    MPI_Comm_size( MPI_COMM_WORLD, &ntasks );

    /* to receive the iteration times, column number sent to it */
    MPI_Recv( &iterTimes, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    MPI_Recv( &colNum, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );

    /* to receive the size of the data block the master will send to it */
    MPI_Recv( &blockSize, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );

    rowNum = blockSize/colNum;

    /* Allocate memory for the data block and a swap block it will receive */
    work = ( xel * ) calloc( blockSize, sizeof( xel ) ); 
    swapBlock = ( xel * ) calloc( blockSize, sizeof( xel ) ); 
    tempP = work;
    lastRow = work+blockSize-colNum;

    /* to receive the data block */
    MPI_Recv( work, blockSize*sizeof( xel ), MPI_BYTE,
      0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );

    /* memcpy( swapBlock, work, blockSize*sizeof( xel ) ); */

    upBound = ( xel * ) calloc( colNum, sizeof( xel ) );
    lowBound = ( xel * ) calloc( colNum, sizeof( xel ) );

    for ( i=0; i<iterTimes; i++ ) {
      /* Before computing begins, exchange data with
       * neighborhood
       */
      /* send the upper bound data to its upper neighbor */
      if( myRank>1 ) {
        MPI_Send( tempP, colNum*sizeof( xel ), MPI_BYTE,
          myRank-1, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
      }

      /*
       * get the lower bound data from its lower neighbor 
       * if the node is last one, just copy the last row
       * to its lower bound area
       */
      if( myRank==ntasks-1 ) {
        memcpy( lowBound, lastRow, colNum*sizeof( xel ) );
      } else {
        MPI_Recv( lowBound, colNum*sizeof( xel ), MPI_BYTE,
          myRank+1, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
        MPI_Send( lastRow, colNum*sizeof( xel ), MPI_BYTE,
          myRank+1, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
      }

      if( myRank>1 ) {
        MPI_Recv( upBound, colNum*sizeof( xel ), MPI_BYTE,
          myRank-1, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
      } else {
        memcpy( upBound, work, colNum*sizeof( xel ) );
      }

      /* computing begins here */

      prevRow = upBound; nextRow=work+culNum;
      temp0=work; temp1=swapBlock;

      while( temp0<work+blockSize; ) {
        for( j=0; j<blockSize-colNum; j++ ) {
          if( j%colNum==0 ) { /* left boundary */
            temp1[ j ] = ( temp0[ j ] + prevRow[ j ]
              + nextRow[ j ] + temp0[j-1] + temp0[ j+1 ] )/5 ; 
          } else if ( (j+1)%colNum==0 ) { /* right boundary */
            temp1[ j ] = ( temp0[ j ] + prevRow[ j ]
              + nextRow[ j ] + temp0[j-1] + temp0[ j+1 ] )/5 ;
          } else {
            swap[ j ] = ( work[ j-colNum ] + work[ j+colNum ]
              +work[ j ] + work[ j-1 ] + work[ j+1 ] )/5;
          }
        }

        prevRow = temp0; temp0=nextRow;

        if(temp0<work+blockSize-colNum)
          nextRow = nextRow+colNum;
        else
          nextRow = lowBound;
      }

        /* swap the pointers to original buffer and
         * newly computed result buffer */
        tempP = work;
        work = swapBlock;
        swapBlock = tempP;
      }
    }

    if( status.MPI_TAG==DIETAG ) {
      MPI_Send( &result, 1, MPI_INT, 0, 0, MPI_COMM_WORLD );
      return;
    }

/* free memory */
}
